/*
 *
 *                  Copyright 1991,1992 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */
/*#include "win_filt.h"*/
#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"
#define __AWK_STRID__
#include "awk.h"

 


#define MAX_BUF_LEN    100


static LPVOID _awk_hash_names        = NIL;  /* pointer to the hash table containing */
                                           /* all 'awk' keyword names. */

static LPVOID _awk_hash_templates    = NIL;  /* pointer to the hash table containing */
                                           /* an abbreviated name and its template */



/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _awk_hash_templates.
 */
typedef struct _LanguageTemplate {         
    LPSTR abbrev;
    LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR awk_INDENT_STR = "\377if\377while\377else\377do\377for\377";

/*
 * _awk_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _awk_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _awk_templates[] = {
    { "at|atan|atan2",       "atan2(&)" },
    { "br|bre|brea|break",          "break;\n" },
    { "cl|clo|clos|close",      "close(&)" },
    { "co|con|cont|conti|contin|continu|continue", "continue;\n" },
    { "cos",            "cos(&)" },
    { "de|del|dele|delet|delete|",  "delete(&[])" },
    { "do",                         "do%0{\n%1&\n%2}\nwhile( );" },
    { "ex|exi|exit",        "exit" },
    { "exp",            "exp(&)" },
    { "fo|for" ,                    "for (&; ; )%0{\n}" },
    { "gs|gsu|gsub",        "gsub(&)" },
    { "if",                         "if (&)\n\t{" },
    { "in|ind|inde|index",      "index(&,)" },
    { "int",            "int(&)" },
    { "le|len|leng|lengt|length",   "length(&)" },
    { "lo|log",         "log(&)" },
    { "ma|mat|matc|match",      "match(&,)" },
    { "ne|nex|next",        "next" },
    { "ra|ran|rand",        "rand(&)" },
    { "re|ret|retu|retur|return",   "return(&);\n" },
    { "si|sin",         "sin(&)" },
    { "sp|spl|spli|split",      "split(&,,)" },
    { "sq|sqr|sqrt",        "sqrt(&)" },
    { "sra|sran|srand",     "srand(&)" },
    { "su|sub",         "sub(&,,)" },
    { "subs|subst|substr",      "substr(&,,)" },
    { "sy|sys|syst|syste|system",   "system(&)" },
    { "tol|tolo|tolow|tolowe|tolower", "tolower(&)" },
    { "tou|toup|toupp|touppe|toupper", "toupper(&)" },
    { "wh|whi|whil|while",          "while (&)%0{\n}" },
    { "{",                          "%0{\n%1&\n%2}" },
    { NIL,       NIL },
   
};
    /* Define the list of 'awk' keyword names to appear in
 * the keyword color if language coloring is enabled. Included are commands 
 * that may be available in NAWK and GAWK versions of the language.
 */
LPSTR _awk_names[] = {
       "atan2", "break",   "close",   "continue",   "cos",
       "delete",  "do",   "exit",   "exp",   "else",   "for",
       "function", "getline", "goto",   "gsub",   "if",   "index",
       "int","length",  "log",   "match",   "next",   "print",   "printf",
       "rand",  "return",  "sin",   "split",   "sprintf",   "sqrt",
       "srand",   "sub",   "substr",  "system",   "tolower",   "toupper",
       "while", "BEGIN", "END", "FNR","FILENAME", "FS", "NF", "NR", "OFS", "ORS", "RS",
    NIL,
};



static long _awk_in_comment( long offset );

void DLL _awk_init( void );
BOOL DLL _awk_expand( void );
long DLL _awk_language_colors_update( long firstline, long lastline );
void DLL _awk_language_colors( long numlines );
void DLL _awk_assign_template( LPSTR abbrev, LPSTR template );
void DLL _awk_add_keyword( LPSTR keyword );
BOOL DLL _awk_indent( void );

LIBMAIN    /* defines the LibMain() */

void DLL 
_init()
{
    _awk_init();
}

/*
 ** _awk_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_awk_init( void )
{
    static BOOL firstTime = TRUE;
    int    i;

    if (firstTime)
    {
        LibExport( "BOOL _awk_expand");
        LibExport( "_awk_assign_template LPSTR LPSTR");
        LibExport( "_awk_add_keyword LPSTR");
        LibExport( "BOOL _awk_indent");
        LibExport( "_awk_language_colors_update long long");
        LibExport( "_awk_language_colors long");

        firstTime = FALSE;

        /* Create a hash table to hold all of the 'awk' keyword names.
         */
        _awk_hash_names = HashCreateTable( 401 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _awk_names[ i ] != NIL )
        {
            LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _awk_hash_names, 
                _awk_names[i], strlen( _awk_names[i] ) );
            if (lpf)
            {
                *lpf = StrNew( _awk_names[i] );
                i++;
            }
        }

        /* Create a hash table to hold all of the 'awk' template names.
         */
        _awk_hash_templates = HashCreateTable( 301 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _awk_templates[ i ].abbrev != NIL )
        {
            _awk_assign_template( _awk_templates[i].abbrev, _awk_templates[i].template );
            i++;
        }
    }
}




/*
 ** _awk_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'awk' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_awk_expand( void )
{
    LPVOID htEntry;
    LPSTR  _awk_string = NULL;
    LPSTR  str;
	BOOL   tabUsage;

    MarkSavePos();

	tabUsage = BufQTabUsage();
    BufSetTabUsage( FALSE );

    /* Check to see if we are currently positioned inside a 'awk' comment
     */
    if (_awk_in_comment( -1L ) != -1)
            goto _awk_expand_incomplete;

    str = BufReadStr( 65535 );     /* read entire line */

    _awk_string = StrTrim( str, " \t" );
    StrFree( str );

    if (!_awk_string)
        goto _awk_expand_incomplete;

    /* Search for a keyword match in the hash table.
     */
    htEntry = HashFindEntry( _awk_hash_templates, _awk_string, strlen( _awk_string ) );
    StrFree( _awk_string );

    if (htEntry)
    {
        /* A template exists for the current keyword, delete the keyword
         * and insert the template.
         */
        MarkDropPos( 1 );
        str = *(LPSTR*)htEntry;
        BufDelToEOL();
        _ext_expand_template( str );

    } 
    else
        goto _awk_expand_incomplete;
    BufSetTabUsage( tabUsage );
    return TRUE;

_awk_expand_incomplete:
    MarkRestorePos();
    BufSetTabUsage( tabUsage );
	
    return FALSE;
}

/*
 ** _awk_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_awk_language_colors_update( long firstline, long lastline )
{
    long startoff;
	long count = 0;
    
    if (firstline && lastline)
    {
        /* Begin coloring at the previous comment if
         * currently located in the middle of one.
         */
        _PosInit( 0 );
        if (firstline == 1)
            _PosPrevLine( 0L );
        else
            _PosNextLine( firstline - 1);
        startoff = _PosQOffset();

        /* Clear all attributes for the line range and
         * the reset them within _awk_language_colors.
         */
        AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
        _PosInit( startoff );
		count = lastline - firstline + 1;
        _awk_language_colors( count );
    }
    return count;
    
}

/*
 ** _awk_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_awk_language_colors( long numlines )

{
    long  startoff;
    long  startline;
    long  endoff;
    long  inc;
    WORD  ch;
    LPSTR buf;
    int   i = 0;
    int   comment_color = ColorComments( -1 );
    int   keyword_color = ColorKeywords( -1 );
    long  maxlines = ExtCommentSearchLimit( -1 );
    int   quote_color      = ColorAlternate1( -1 );
    int   number_color     = ColorAlternate2( -1 );
  /*    int   brace_color      = ColorAlternate3( -1 );*/
  /*    int   preproc_color    = ColorAlternate4( -1 );*/


    buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

    /* Start at the top of the buffer or current position 
     */
    if (numlines == -1)
    {
        _PosInit( 0 );  
        numlines = 0x7FFFFFFF;
    }
    else
    {
        startoff  = _PosQOffset();
        startline = _PosQLine();
        endoff    = _awk_in_comment( startoff );
        if (endoff == -1)
            _PosInit( startoff );
        else
        {
            _PosInit( endoff );
            numlines += (startline - _PosQLine());
        }
    }
    ch = _PosCurrentChar();
    startoff = _PosQOffset();

    maxlines = max( numlines, maxlines );

    while ((ch != EOF_CHAR) && (numlines > 0))
    {
        if (ch == '\n')
            numlines--;

        //Quoted
        if (ch == '"')
        {
            UINT quote = ch;

            startoff = _PosQOffset();

            do 
            {
                ch = _PosNextChar();
                if (ch == '\\')
                {
                    ch = _PosNextChar();
                    if ((ch == quote) || (ch == '\\'))
                        ch = _PosNextChar();
                }
                if ((ch == EOF_CHAR) || (ch == '\n'))
                    break;
            } while (ch != quote);

            if (ch == quote)
                ch = _PosNextChar();

            /* color the string */
            if (quote_color)
            {
                endoff = _PosQOffset();
                _PosInit( startoff );
                _PosSetColor( quote_color, endoff - startoff );
                _PosInit( endoff );
            }
            continue;
        }
        //Digit
        else if (number_color && isdigit( (BYTE)ch ) )  {
            int j = 0;
          
            UINT lastch = _PosPrevChar();
          
          if (lastch != EOF_CHAR)
                _PosNextChar();

            startoff = _PosQOffset();
            do
            {
                j++;
                ch = _PosNextChar();
            } while (isdigit( (BYTE)ch ) || (ch == '.') || (ch == 'x' && j == 1));

            if ((  lastch != ' ' 
                && lastch != '\t' 
                && lastch != '\r' 
                && lastch != '\n' 
                && lastch != EOF_CHAR
                && !ispunct( (BYTE)lastch ))
                || lastch == '_'
                || (ch == '_'))
                continue;
           
            if (ch == 'l' || ch == 'L')
                ch = _PosNextChar();

            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( number_color, endoff - startoff );
            _PosInit( endoff );
            continue;
        }
        //Alpha
        else if (isalpha( (BYTE)ch ) || (ch == '_') )
        {
            UINT lastch = _PosPrevChar();

            if (lastch != EOF_CHAR)
                _PosNextChar();

            i = 0;
            startoff = _PosQOffset();

            do
            {
                if (i < MAX_BUF_LEN)
                    buf[i++] = (char)ch;
                ch = _PosNextChar();
            } while (isalpha( (BYTE)ch ) || (ch == '_')|| (ch=='2'));

            if ((lastch != ' ' 
                        && lastch != '\t' 
                        && lastch != '\r' 
                        && lastch != '\n' 
                        && lastch != EOF_CHAR
                        && !ispunct( (BYTE)lastch ))
                    || (ch == '_'))
                i = 0;
            continue;
        }
        else if (i)
        {
            if (HashFindEntry( _awk_hash_names, _fstrupr(buf), i ))
            {
                endoff = _PosQOffset();
                _PosInit( startoff );
                _PosSetColor( keyword_color, endoff - startoff );
                _PosInit( endoff );
            }
            i = 0;
        } 

        //Comments
        if (ch == '#')
        {
            startoff = _PosQOffset();

            inc = _PosNextLine( 1 );
            numlines--;
            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( comment_color, endoff - startoff - inc );
            _PosInit( endoff );
            ch = _PosCurrentChar();
            continue;
        }

        ch = _PosNextChar();
    }
    MemFree( buf );
}




void DLL
_awk_add_keyword( LPSTR keyword )
{
    if (keyword && *keyword)
    {
        LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _awk_hash_names, 
                keyword, strlen( keyword ) );
        if (lpf)
            *lpf = StrNew( keyword );
    }
}



/*
 ** _awk_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'awk' template strings
 *    without changing the code.
 */
void DLL
_awk_assign_template( LPSTR abbrev, LPSTR template )
{
    _ext_assign_template( _awk_hash_templates, abbrev, template );
}


/*
 ** _awk_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'awk' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in awk_INDENT_STR - perform smart
 *       indentation.  
 */
BOOL DLL
_awk_indent( void )
{
    long   curline;
    long   col;
    LPSTR  _awk_string;
    LPSTR  str;
    long   matchpos;
    
    MarkSavePos();

    curline = BufQCurrentLine();

    /* Check to see if we are currently positioned inside a 'awk' comment
     */
    if (_awk_in_comment( -1L ) != -1)
        goto _awk_indent_incomplete;

    /* Not in a comment so indent according to the current 'awk' construct.
     */
    if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
            || (BufQCurrentLine() != curline) )
        goto _awk_indent_incomplete;

    /* Read the keyword at the beginning of the current line.
     */
    col = BufQCurrentCol();
    SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
    str = BufReadStr( col - BufQCurrentCol() );
    _awk_string = StrTrim( str, " \t" );
    StrFree( str );

    /* Set prefix and suffix to \377 to perform a quick lookup of the name
     */
    str = StrNew( "\377" );
    _awk_string = StrAppend( _awk_string, str );
    str = StrAppend( str, _awk_string );
    StrFree( _awk_string );
    
    matchpos = StrMatch( str, awk_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
    StrFree( str );

    if (!matchpos)
        /* Not a 'awk' keyword, do normal indentation. */
        goto _awk_indent_incomplete;
        
    MarkRestorePos();

    _ext_insert_indented_EOL();
    BufInsertChar( '\t' );
    return TRUE;

    _awk_indent_incomplete:
        MarkRestorePos();

    return FALSE;
}



/*--  Private functions --*/


/*
 ** _awk_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'awk' comment by search for a # in column 1.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_awk_in_comment( long offset )
{
    
    WORD  ch;
    /*offset = 1;*/
    _PosInit( offset );
    ch = _PosCurrentChar();

    do
    {
        if (ch == '\n')
            return -1;
        else if (ch == '#')
            return _PosQOffset();  /* currrent line is in a comment */
        ch = _PosPrevChar();
    } while (ch != EOF_CHAR);

    return -1L;
}

